@@ -1,17 +1,15 @@
-from __future__ import absolute_import
-import os
-import re
-import base64
+from __future__ import absolute_import, print_function, division
 
 import configargparse
-
-from netlib.tcp import Address, sslversion_choices
-import netlib.utils
-from . import filt, utils, version
-from .proxy import config
-
-APP_HOST = "mitm.it"
-APP_PORT = 80
+import os
+import re
+from mitmproxy import exceptions
+from mitmproxy import flowfilter
+from mitmproxy import options
+from mitmproxy import platform
+from netlib import human
+from netlib import tcp
+from netlib import version
 
 
 class ParseException(Exception):
@@ -34,7 +32,7 @@ def _parse_hook(s):
     if not a:
         raise ParseException("Empty clause: %s" % str(patt))
 
-    if not filt.parse(patt):
+    if not flowfilter.parse(patt):
         raise ParseException("Malformed filter pattern: %s" % patt)
 
     return patt, a, b
@@ -70,7 +68,7 @@ def parse_replace_hook(s):
     try:
         re.compile(regex)
     except re.error as e:
-        raise ParseException("Malformed replacement regex: %s" % str(e.message))
+        raise ParseException("Malformed replacement regex: %s" % str(e))
     return patt, regex, replacement
 
 
@@ -103,96 +101,169 @@ def parse_setheader(s):
     return _parse_hook(s)
 
 
-def parse_server_spec(url):
-    try:
-        p = netlib.utils.parse_url(url)
-        if p[0] not in ("http", "https"):
-            raise ValueError()
-    except ValueError:
-        raise configargparse.ArgumentTypeError(
-            "Invalid server specification: %s" % url
-        )
-
-    address = Address(p[1:3])
-    scheme = p[0].lower()
-    return config.ServerSpec(scheme, address)
-
-
-def parse_upstream_auth(auth):
-    pattern = re.compile(".+:")
-    if pattern.search(auth) is None:
-        raise configargparse.ArgumentTypeError(
-            "Invalid upstream auth specification: %s" % auth
-        )
-    return "Basic" + " " + base64.b64encode(auth)
-
-
-def get_common_options(options):
+def get_common_options(args):
     stickycookie, stickyauth = None, None
-    if options.stickycookie_filt:
-        stickycookie = options.stickycookie_filt
+    if args.stickycookie_filt:
+        stickycookie = args.stickycookie_filt
 
-    if options.stickyauth_filt:
-        stickyauth = options.stickyauth_filt
+    if args.stickyauth_filt:
+        stickyauth = args.stickyauth_filt
 
-    stream_large_bodies = utils.parse_size(options.stream_large_bodies)
+    stream_large_bodies = args.stream_large_bodies
+    if stream_large_bodies:
+        stream_large_bodies = human.parse_size(stream_large_bodies)
 
     reps = []
-    for i in options.replace:
+    for i in args.replace:
         try:
             p = parse_replace_hook(i)
         except ParseException as e:
-            raise configargparse.ArgumentTypeError(e.message)
+            raise exceptions.OptionsError(e)
         reps.append(p)
-    for i in options.replace_file:
+    for i in args.replace_file:
         try:
             patt, rex, path = parse_replace_hook(i)
         except ParseException as e:
-            raise configargparse.ArgumentTypeError(e.message)
+            raise exceptions.OptionsError(e)
         try:
             v = open(path, "rb").read()
         except IOError as e:
-            raise configargparse.ArgumentTypeError(
+            raise exceptions.OptionsError(
                 "Could not read replace file: %s" % path
             )
         reps.append((patt, rex, v))
 
     setheaders = []
-    for i in options.setheader:
+    for i in args.setheader:
         try:
             p = parse_setheader(i)
         except ParseException as e:
-            raise configargparse.ArgumentTypeError(e.message)
+            raise exceptions.OptionsError(e)
         setheaders.append(p)
 
-    return dict(
-        app=options.app,
-        app_host=options.app_host,
-        app_port=options.app_port,
+    if args.outfile and args.outfile[0] == args.rfile:
+        if args.outfile[1] == "wb":
+            raise exceptions.OptionsError(
+                "Cannot use '{}' for both reading and writing flows. "
+                "Are you looking for --afile?".format(args.rfile)
+            )
+        else:
+            raise exceptions.OptionsError(
+                "Cannot use '{}' for both reading and appending flows. "
+                "That would trigger an infinite loop."
+            )
 
-        anticache=options.anticache,
-        anticomp=options.anticomp,
-        client_replay=options.client_replay,
-        kill=options.kill,
-        no_server=options.no_server,
-        refresh_server_playback=not options.norefresh,
-        rheaders=options.rheaders,
-        rfile=options.rfile,
+    # Proxy config
+    certs = []
+    for i in args.certs:
+        parts = i.split("=", 1)
+        if len(parts) == 1:
+            parts = ["*", parts[0]]
+        certs.append(parts)
+
+    body_size_limit = args.body_size_limit
+    if body_size_limit:
+        try:
+            body_size_limit = human.parse_size(body_size_limit)
+        except ValueError as e:
+            raise exceptions.OptionsError(
+                "Invalid body size limit specification: %s" % body_size_limit
+            )
+
+    # Establish proxy mode
+    c = 0
+    mode, upstream_server = "regular", None
+    if args.transparent_proxy:
+        c += 1
+        if not platform.resolver:
+            raise exceptions.OptionsError(
+                "Transparent mode not supported on this platform."
+            )
+        mode = "transparent"
+    if args.socks_proxy:
+        c += 1
+        mode = "socks5"
+    if args.reverse_proxy:
+        c += 1
+        mode = "reverse"
+        upstream_server = args.reverse_proxy
+    if args.upstream_proxy:
+        c += 1
+        mode = "upstream"
+        upstream_server = args.upstream_proxy
+    if c > 1:
+        raise exceptions.OptionsError(
+            "Transparent, SOCKS5, reverse and upstream proxy mode "
+            "are mutually exclusive. Read the docs on proxy modes "
+            "to understand why."
+        )
+    if args.add_upstream_certs_to_client_chain and args.no_upstream_cert:
+        raise exceptions.OptionsError(
+            "The no-upstream-cert and add-upstream-certs-to-client-chain "
+            "options are mutually exclusive. If no-upstream-cert is enabled "
+            "then the upstream certificate is not retrieved before generating "
+            "the client certificate chain."
+        )
+
+    if args.quiet:
+        args.verbose = 0
+
+    return dict(
+        app=args.app,
+        app_host=args.app_host,
+        app_port=args.app_port,
+
+        anticache=args.anticache,
+        anticomp=args.anticomp,
+        client_replay=args.client_replay,
+        replay_kill_extra=args.replay_kill_extra,
+        no_server=args.no_server,
+        refresh_server_playback=not args.norefresh,
+        server_replay_use_headers=args.server_replay_use_headers,
+        rfile=args.rfile,
         replacements=reps,
         setheaders=setheaders,
-        server_replay=options.server_replay,
-        scripts=options.scripts,
+        server_replay=args.server_replay,
+        scripts=args.scripts,
         stickycookie=stickycookie,
         stickyauth=stickyauth,
         stream_large_bodies=stream_large_bodies,
-        showhost=options.showhost,
-        outfile=options.outfile,
-        verbosity=options.verbose,
-        nopop=options.nopop,
-        replay_ignore_content=options.replay_ignore_content,
-        replay_ignore_params=options.replay_ignore_params,
-        replay_ignore_payload_params=options.replay_ignore_payload_params,
-        replay_ignore_host=options.replay_ignore_host
+        showhost=args.showhost,
+        outfile=args.outfile,
+        verbosity=args.verbose,
+        server_replay_nopop=args.server_replay_nopop,
+        server_replay_ignore_content=args.server_replay_ignore_content,
+        server_replay_ignore_params=args.server_replay_ignore_params,
+        server_replay_ignore_payload_params=args.server_replay_ignore_payload_params,
+        server_replay_ignore_host=args.server_replay_ignore_host,
+
+        auth_nonanonymous = args.auth_nonanonymous,
+        auth_singleuser = args.auth_singleuser,
+        auth_htpasswd = args.auth_htpasswd,
+        add_upstream_certs_to_client_chain = args.add_upstream_certs_to_client_chain,
+        body_size_limit = body_size_limit,
+        cadir = args.cadir,
+        certs = certs,
+        ciphers_client = args.ciphers_client,
+        ciphers_server = args.ciphers_server,
+        clientcerts = args.clientcerts,
+        http2 = args.http2,
+        ignore_hosts = args.ignore_hosts,
+        listen_host = args.addr,
+        listen_port = args.port,
+        mode = mode,
+        no_upstream_cert = args.no_upstream_cert,
+        spoof_source_address = args.spoof_source_address,
+        rawtcp = args.rawtcp,
+        websockets = args.websockets,
+        upstream_server = upstream_server,
+        upstream_auth = args.upstream_auth,
+        ssl_version_client = args.ssl_version_client,
+        ssl_version_server = args.ssl_version_server,
+        ssl_insecure = args.ssl_insecure,
+        ssl_verify_upstream_trusted_cadir = args.ssl_verify_upstream_trusted_cadir,
+        ssl_verify_upstream_trusted_ca = args.ssl_verify_upstream_trusted_ca,
+        tcp_hosts = args.tcp_hosts,
     )
 
 
@@ -202,6 +273,11 @@ def basic_options(parser):
         action='version',
         version="%(prog)s" + " " + version.VERSION
     )
+    parser.add_argument(
+        '--sysinfo',
+        action='store_true',
+        dest='sysinfo',
+    )
     parser.add_argument(
         '--shortversion',
         action='version',
@@ -219,8 +295,8 @@ def basic_options(parser):
     )
     parser.add_argument(
         "--cadir",
-        action="store", type=str, dest="cadir", default=config.CA_DIR,
-        help="Location of the default mitmproxy CA files. (%s)" % config.CA_DIR
+        action="store", type=str, dest="cadir", default=options.CA_DIR,
+        help="Location of the default mitmproxy CA files. (%s)" % options.CA_DIR
     )
     parser.add_argument(
         "--host",
@@ -261,8 +337,8 @@ def basic_options(parser):
     )
     parser.add_argument(
         "-v", "--verbose",
-        action="store_const", dest="verbose", default=1, const=2,
-        help="Increase event log verbosity."
+        action="store_const", dest="verbose", default=2, const=3,
+        help="Increase log verbosity."
     )
     outfile = parser.add_mutually_exclusive_group()
     outfile.add_argument(
@@ -300,16 +376,18 @@ def basic_options(parser):
 
 
 def proxy_modes(parser):
-    group = parser.add_argument_group("Proxy Modes").add_mutually_exclusive_group()
+    group = parser.add_argument_group("Proxy Modes")
     group.add_argument(
         "-R", "--reverse",
         action="store",
-        type=parse_server_spec,
+        type=str,
         dest="reverse_proxy",
         default=None,
         help="""
             Forward all requests to upstream HTTP server:
-            http[s][2http[s]]://host[:port]
+            http[s]://host[:port]. Clients can always connect both
+            via HTTPS and HTTP, the connection to the server is
+            determined by the specified scheme.
         """
     )
     group.add_argument(
@@ -325,7 +403,7 @@ def proxy_modes(parser):
     group.add_argument(
         "-U", "--upstream",
         action="store",
-        type=parse_server_spec,
+        type=str,
         dest="upstream_proxy",
         default=None,
         help="Forward all requests to upstream proxy server: http://host[:port]"
@@ -359,7 +437,7 @@ def proxy_options(parser):
         help="""
             Generic TCP SSL proxy mode for all hosts that match the pattern.
             Similar to --ignore, but SSL connections are intercepted. The
-            communication contents are printed to the event log in verbose mode.
+            communication contents are printed to the log in verbose mode.
         """
     )
     group.add_argument(
@@ -369,7 +447,7 @@ def proxy_options(parser):
     )
     group.add_argument(
         "-p", "--port",
-        action="store", type=int, dest="port", default=8080,
+        action="store", type=int, dest="port", default=options.LISTEN_PORT,
         help="Proxy service port."
     )
     group.add_argument(
@@ -383,7 +461,7 @@ def proxy_options(parser):
     parser.add_argument(
         "--upstream-auth",
         action="store", dest="upstream_auth", default=None,
-        type=parse_upstream_auth,
+        type=str,
         help="""
             Proxy Authentication:
             username:password
@@ -396,6 +474,18 @@ def proxy_options(parser):
                         "Disabled by default. "
                         "Default value will change in a future version."
                         )
+    websockets = group.add_mutually_exclusive_group()
+    websockets.add_argument("--websockets", action="store_true", dest="websockets")
+    websockets.add_argument("--no-websockets", action="store_false", dest="websockets",
+                            help="Explicitly enable/disable experimental WebSocket support. "
+                                 "Disabled by default as messages are only printed to the event log and not retained. "
+                                 "Default value will change in a future version."
+                            )
+    group.add_argument(
+        "--spoof-source-address",
+        action="store_true", dest="spoof_source_address",
+        help="Use the client's IP for server-side connections"
+    )
 
 
 def proxy_ssl_options(parser):
@@ -416,7 +506,7 @@ def proxy_ssl_options(parser):
              'as the first entry. Can be passed multiple times.')
     group.add_argument(
         "--ciphers-client", action="store",
-        type=str, dest="ciphers_client", default=config.DEFAULT_CLIENT_CIPHERS,
+        type=str, dest="ciphers_client", default=options.DEFAULT_CLIENT_CIPHERS,
         help="Set supported ciphers for client connections. (OpenSSL Syntax)"
     )
     group.add_argument(
@@ -441,10 +531,9 @@ def proxy_ssl_options(parser):
              "that will be served to the proxy client, as extras."
     )
     group.add_argument(
-        "--verify-upstream-cert", default=False,
-        action="store_true", dest="ssl_verify_upstream_cert",
-        help="Verify upstream server SSL/TLS certificates and fail if invalid "
-             "or not present."
+        "--insecure", default=False,
+        action="store_true", dest="ssl_insecure",
+        help="Do not verify upstream server SSL/TLS certificates."
     )
     group.add_argument(
         "--upstream-trusted-cadir", default=None, action="store",
@@ -460,14 +549,14 @@ def proxy_ssl_options(parser):
     group.add_argument(
         "--ssl-version-client", dest="ssl_version_client",
         default="secure", action="store",
-        choices=sslversion_choices.keys(),
+        choices=tcp.sslversion_choices.keys(),
         help="Set supported SSL/TLS versions for client connections. "
              "SSLv2, SSLv3 and 'all' are INSECURE. Defaults to secure, which is TLS1.0+."
     )
     group.add_argument(
         "--ssl-version-server", dest="ssl_version_server",
         default="secure", action="store",
-        choices=sslversion_choices.keys(),
+        choices=tcp.sslversion_choices.keys(),
         help="Set supported SSL/TLS versions for server connections. "
              "SSLv2, SSLv3 and 'all' are INSECURE. Defaults to secure, which is TLS1.0+."
     )
@@ -482,18 +571,18 @@ def onboarding_app(parser):
     )
     group.add_argument(
         "--app-host",
-        action="store", dest="app_host", default=APP_HOST, metavar="host",
+        action="store", dest="app_host", default=options.APP_HOST, metavar="host",
         help="""
             Domain to serve the onboarding app from. For transparent mode, use
             an IP when a DNS entry for the app domain is not present. Default:
             %s
-        """ % APP_HOST
+        """ % options.APP_HOST
     )
     group.add_argument(
         "--app-port",
         action="store",
         dest="app_port",
-        default=APP_PORT,
+        default=options.APP_PORT,
         type=int,
         metavar="80",
         help="Port to serve the onboarding app from."
@@ -517,13 +606,13 @@ def server_replay(parser):
         help="Replay server responses from a saved file."
     )
     group.add_argument(
-        "-k", "--kill",
-        action="store_true", dest="kill", default=False,
+        "-k", "--replay-kill-extra",
+        action="store_true", dest="replay_kill_extra", default=False,
         help="Kill extra requests during replay."
     )
     group.add_argument(
-        "--rheader",
-        action="append", dest="rheaders", type=str,
+        "--server-replay-use-header",
+        action="append", dest="server_replay_use_headers", type=str,
         help="Request headers to be considered during replay. "
              "Can be passed multiple times."
     )
@@ -537,21 +626,21 @@ def server_replay(parser):
     )
     group.add_argument(
         "--no-pop",
-        action="store_true", dest="nopop", default=False,
+        action="store_true", dest="server_replay_nopop", default=False,
         help="Disable response pop from response flow. "
              "This makes it possible to replay same response multiple times."
     )
     payload = group.add_mutually_exclusive_group()
     payload.add_argument(
         "--replay-ignore-content",
-        action="store_true", dest="replay_ignore_content", default=False,
+        action="store_true", dest="server_replay_ignore_content", default=False,
         help="""
             Ignore request's content while searching for a saved flow to replay
         """
     )
     payload.add_argument(
         "--replay-ignore-payload-param",
-        action="append", dest="replay_ignore_payload_params", type=str,
+        action="append", dest="server_replay_ignore_payload_params", type=str,
         help="""
             Request's payload parameters (application/x-www-form-urlencoded or multipart/form-data) to
             be ignored while searching for a saved flow to replay.
@@ -561,7 +650,7 @@ def server_replay(parser):
 
     group.add_argument(
         "--replay-ignore-param",
-        action="append", dest="replay_ignore_params", type=str,
+        action="append", dest="server_replay_ignore_params", type=str,
         help="""
             Request's parameters to be ignored while searching for a saved flow
             to replay. Can be passed multiple times.
@@ -570,7 +659,7 @@ def server_replay(parser):
     group.add_argument(
         "--replay-ignore-host",
         action="store_true",
-        dest="replay_ignore_host",
+        dest="server_replay_ignore_host",
         default=False,
         help="Ignore request's destination host while searching for a saved flow to replay")
 
@@ -671,8 +760,8 @@ def mitmproxy():
         usage="%(prog)s [options]",
         args_for_setting_config_path=["--conf"],
         default_config_files=[
-            os.path.join(config.CA_DIR, "common.conf"),
-            os.path.join(config.CA_DIR, "mitmproxy.conf")
+            os.path.join(options.CA_DIR, "common.conf"),
+            os.path.join(options.CA_DIR, "mitmproxy.conf")
         ],
         add_config_file_help=True,
         add_env_var_help=True
@@ -695,7 +784,7 @@ def mitmproxy():
         help="Show event log."
     )
     parser.add_argument(
-        "-f", "--follow",
+        "--follow",
         action="store_true", dest="follow",
         help="Follow flow list."
     )
@@ -714,9 +803,9 @@ def mitmproxy():
         help="Intercept filter expression."
     )
     group.add_argument(
-        "-l", "--limit", action="store",
-        type=str, dest="limit", default=None,
-        help="Limit filter expression."
+        "-f", "--filter", action="store",
+        type=str, dest="filter", default=None,
+        help="Filter view expression."
     )
     return parser
 
@@ -726,8 +815,8 @@ def mitmdump():
         usage="%(prog)s [options] [filter]",
         args_for_setting_config_path=["--conf"],
         default_config_files=[
-            os.path.join(config.CA_DIR, "common.conf"),
-            os.path.join(config.CA_DIR, "mitmdump.conf")
+            os.path.join(options.CA_DIR, "common.conf"),
+            os.path.join(options.CA_DIR, "mitmdump.conf")
         ],
         add_config_file_help=True,
         add_env_var_help=True
@@ -756,8 +845,8 @@ def mitmweb():
         usage="%(prog)s [options]",
         args_for_setting_config_path=["--conf"],
         default_config_files=[
-            os.path.join(config.CA_DIR, "common.conf"),
-            os.path.join(config.CA_DIR, "mitmweb.conf")
+            os.path.join(options.CA_DIR, "common.conf"),
+            os.path.join(options.CA_DIR, "mitmweb.conf")
         ],
         add_config_file_help=True,
         add_env_var_help=True
