@@ -1,11 +1,13 @@
+from __future__ import absolute_import, print_function, division
+
 import blinker
 
 # Show a status message in the action bar
-sig_add_event = blinker.Signal()
+sig_add_log = blinker.Signal()
 
 
-def add_event(e, level):
-    sig_add_event.send(
+def add_log(e, level):
+    sig_add_log.send(
         None,
         e=e,
         level=level
@@ -41,3 +43,4 @@ flowlist_change = blinker.Signal()
 # Pop and push view state onto a stack
 pop_view_state = blinker.Signal()
 push_view_state = blinker.Signal()
+replace_view_state = blinker.Signal()
