@@ -1,16 +1,25 @@
-from __future__ import (absolute_import, print_function, division)
+from __future__ import absolute_import, print_function, division
 
-from .http import (
-    HTTPFlow, HTTPRequest, HTTPResponse, Headers, decoded,
-    make_error_response, make_connect_request, make_connect_response, expect_continue_response
-)
+from netlib.http import decoded
 from .connections import ClientConnection, ServerConnection
 from .flow import Flow, Error
+from .http import (
+    HTTPFlow, HTTPRequest, HTTPResponse,
+    make_error_response, make_connect_request, make_connect_response, expect_continue_response
+)
+from .tcp import TCPFlow
+
+FLOW_TYPES = dict(
+    http=HTTPFlow,
+    tcp=TCPFlow,
+)
 
 __all__ = [
-    "HTTPFlow", "HTTPRequest", "HTTPResponse", "Headers", "decoded",
+    "HTTPFlow", "HTTPRequest", "HTTPResponse", "decoded",
     "make_error_response", "make_connect_request",
     "make_connect_response", "expect_continue_response",
     "ClientConnection", "ServerConnection",
     "Flow", "Error",
+    "TCPFlow",
+    "FLOW_TYPES",
 ]
