@@ -1,5 +1,6 @@
 import subprocess
-import pf
+
+from . import pf
 
 """
     Doing this the "right" way by using DIOCNATLOOK on the pf device turns out
@@ -22,15 +23,15 @@ class Resolver(object):
         try:
             stxt = subprocess.check_output(self.STATECMD, stderr=subprocess.STDOUT)
         except subprocess.CalledProcessError as e:
-            if "sudo: a password is required" in e.output:
+            if "sudo: a password is required" in e.output.decode(errors="replace"):
                 insufficient_priv = True
             else:
                 raise RuntimeError("Error getting pfctl state: " + repr(e))
         else:
-            insufficient_priv = "sudo: a password is required" in stxt
+            insufficient_priv = "sudo: a password is required" in stxt.decode(errors="replace")
 
         if insufficient_priv:
             raise RuntimeError(
                 "Insufficient privileges to access pfctl. "
-                "See http://mitmproxy.org/doc/transparent/osx.html for details.")
+                "See http://docs.mitmproxy.org/en/latest/transparent/osx.html for details.")
         return pf.lookup(peer[0], peer[1], stxt)
