@@ -1,47 +1,38 @@
-from __future__ import (absolute_import, print_function, division)
+from __future__ import absolute_import, print_function, division
 
-import sys
-import traceback
+import h2.exceptions
+import netlib.exceptions
 import six
-
+import sys
+import time
+import traceback
+from mitmproxy import exceptions
+from mitmproxy import models
+from mitmproxy.protocol import base
+from mitmproxy.protocol import websockets as pwebsockets
+from netlib import http
 from netlib import tcp
-from netlib.exceptions import HttpException, HttpReadDisconnect, NetlibException
-from netlib.http import Headers
-
-from h2.exceptions import H2Error
-
-from .. import utils
-from ..exceptions import HttpProtocolException, ProtocolException
-from ..models import (
-    HTTPFlow,
-    HTTPResponse,
-    make_error_response,
-    make_connect_response,
-    Error,
-    expect_continue_response
-)
-
-from .base import Layer, Kill
+from netlib import websockets
 
 
-class _HttpTransmissionLayer(Layer):
-
-    def read_request(self):
+class _HttpTransmissionLayer(base.Layer):
+    def read_request_headers(self):
         raise NotImplementedError()
 
     def read_request_body(self, request):
         raise NotImplementedError()
 
+    def read_request(self):
+        request = self.read_request_headers()
+        request.data.content = b"".join(
+            self.read_request_body(request)
+        )
+        request.timestamp_end = time.time()
+        return request
+
     def send_request(self, request):
         raise NotImplementedError()
 
-    def read_response(self, request):
-        response = self.read_response_headers()
-        response.data.content = b"".join(
-            self.read_response_body(request, response)
-        )
-        return response
-
     def read_response_headers(self):
         raise NotImplementedError()
 
@@ -49,11 +40,18 @@ class _HttpTransmissionLayer(Layer):
         raise NotImplementedError()
         yield "this is a generator"  # pragma: no cover
 
+    def read_response(self, request):
+        response = self.read_response_headers()
+        response.data.content = b"".join(
+            self.read_response_body(request, response)
+        )
+        return response
+
     def send_response(self, response):
-        if response.content is None:
-            raise HttpException("Cannot assemble flow with missing content")
+        if response.data.content is None:
+            raise netlib.exceptions.HttpException("Cannot assemble flow with missing content")
         self.send_response_headers(response)
-        self.send_response_body(response, [response.content])
+        self.send_response_body(response, [response.data.content])
 
     def send_response_headers(self, response):
         raise NotImplementedError()
@@ -89,7 +87,7 @@ class ConnectServerConnection(object):
         __nonzero__ = __bool__
 
 
-class UpstreamConnectLayer(Layer):
+class UpstreamConnectLayer(base.Layer):
 
     def __init__(self, ctx, connect_request):
         super(UpstreamConnectLayer, self).__init__(ctx)
@@ -107,7 +105,7 @@ class UpstreamConnectLayer(Layer):
         self.send_request(self.connect_request)
         resp = self.read_response(self.connect_request)
         if resp.status_code != 200:
-            raise ProtocolException("Reconnect: Upstream server refuses CONNECT request")
+            raise exceptions.ProtocolException("Reconnect: Upstream server refuses CONNECT request")
 
     def connect(self):
         if not self.server_conn:
@@ -120,7 +118,7 @@ class UpstreamConnectLayer(Layer):
         if address != self.server_conn.via.address:
             self.ctx.set_server(address)
 
-    def set_server(self, address, server_tls=None, sni=None):
+    def set_server(self, address):
         if self.ctx.server_conn:
             self.ctx.disconnect()
         address = tcp.Address.wrap(address)
@@ -128,13 +126,8 @@ class UpstreamConnectLayer(Layer):
         self.connect_request.port = address.port
         self.server_conn.address = address
 
-        if server_tls:
-            raise ProtocolException(
-                "Cannot upgrade to TLS, no TLS layer on the protocol stack."
-            )
 
-
-class HttpLayer(Layer):
+class HttpLayer(base.Layer):
 
     def __init__(self, ctx, mode):
         super(HttpLayer, self).__init__(ctx)
@@ -146,108 +139,134 @@ class HttpLayer(Layer):
         # We cannot rely on server_conn.tls_established,
         # see https://github.com/mitmproxy/mitmproxy/issues/925
         self.__initial_server_tls = None
+        # Requests happening after CONNECT do not need Proxy-Authorization headers.
+        self.http_authenticated = False
 
     def __call__(self):
         if self.mode == "transparent":
-            self.__initial_server_tls = self._server_tls
+            self.__initial_server_tls = self.server_tls
             self.__initial_server_conn = self.server_conn
         while True:
+            flow = models.HTTPFlow(self.client_conn, self.server_conn, live=self)
             try:
-                request = self.get_request_from_client()
-                self.log("request", "debug", [repr(request)])
-
-                # Handle Proxy Authentication
-                # Proxy Authentication conceptually does not work in transparent mode.
-                # We catch this misconfiguration on startup. Here, we sort out requests
-                # after a successful CONNECT request (which do not need to be validated anymore)
-                if self.mode != "transparent" and not self.authenticate(request):
-                    return
-
+                request = self.get_request_from_client(flow)
                 # Make sure that the incoming request matches our expectations
                 self.validate_request(request)
+            except netlib.exceptions.HttpReadDisconnect:
+                # don't throw an error for disconnects that happen before/between requests.
+                return
+            except netlib.exceptions.HttpException as e:
+                # We optimistically guess there might be an HTTP client on the
+                # other end
+                self.send_error_response(400, repr(e))
+                six.reraise(
+                    exceptions.ProtocolException,
+                    exceptions.ProtocolException(
+                        "HTTP protocol error in client request: {}".format(e)
+                    ),
+                    sys.exc_info()[2]
+                )
 
+            self.log("request", "debug", [repr(request)])
+
+            # Handle Proxy Authentication
+            # Proxy Authentication conceptually does not work in transparent mode.
+            # We catch this misconfiguration on startup. Here, we sort out requests
+            # after a successful CONNECT request (which do not need to be validated anymore)
+            if not (self.http_authenticated or self.authenticate(request)):
+                return
+
+            flow.request = request
+
+            try:
                 # Regular Proxy Mode: Handle CONNECT
                 if self.mode == "regular" and request.first_line_format == "authority":
                     self.handle_regular_mode_connect(request)
                     return
-
-            except HttpReadDisconnect:
-                # don't throw an error for disconnects that happen before/between requests.
+            except (exceptions.ProtocolException, netlib.exceptions.NetlibException) as e:
+                # HTTPS tasting means that ordinary errors like resolution and
+                # connection errors can happen here.
+                self.send_error_response(502, repr(e))
+                flow.error = models.Error(str(e))
+                self.channel.ask("error", flow)
                 return
-            except NetlibException as e:
-                self.send_error_response(400, repr(e))
-                six.reraise(ProtocolException, ProtocolException(
-                    "Error in HTTP connection: %s" % repr(e)), sys.exc_info()[2])
+
+            # update host header in reverse proxy mode
+            if self.config.options.mode == "reverse":
+                if six.PY2:
+                    flow.request.headers["Host"] = self.config.upstream_server.address.host.encode()
+                else:
+                    flow.request.headers["Host"] = self.config.upstream_server.address.host
+
+            # set upstream auth
+            if self.mode == "upstream" and self.config.upstream_auth is not None:
+                flow.request.headers["Proxy-Authorization"] = self.config.upstream_auth
+            self.process_request_hook(flow)
 
             try:
-                flow = HTTPFlow(self.client_conn, self.server_conn, live=self)
-                flow.request = request
-                # set upstream auth
-                if self.mode == "upstream" and self.config.upstream_auth is not None:
-                    self.data.headers["Proxy-Authorization"] = self.config.upstream_auth
-                self.process_request_hook(flow)
+                if websockets.check_handshake(request.headers) and websockets.check_client_version(request.headers):
+                    # We only support RFC6455 with WebSockets version 13
+                    # allow inline scripts to manipulate the client handshake
+                    self.channel.ask("websocket_handshake", flow)
 
                 if not flow.response:
-                    self.establish_server_connection(flow)
+                    self.establish_server_connection(
+                        flow.request.host,
+                        flow.request.port,
+                        flow.request.scheme
+                    )
                     self.get_response_from_server(flow)
                 else:
                     # response was set by an inline script.
                     # we now need to emulate the responseheaders hook.
-                    flow = self.channel.ask("responseheaders", flow)
-                    if flow == Kill:
-                        raise Kill()
+                    self.channel.ask("responseheaders", flow)
 
                 self.log("response", "debug", [repr(flow.response)])
-                flow = self.channel.ask("response", flow)
-                if flow == Kill:
-                    raise Kill()
+                self.channel.ask("response", flow)
                 self.send_response_to_client(flow)
 
                 if self.check_close_connection(flow):
                     return
 
                 # Handle 101 Switching Protocols
-                # It may be useful to pass additional args (such as the upgrade header)
-                # to next_layer in the future
                 if flow.response.status_code == 101:
-                    layer = self.ctx.next_layer(self)
-                    layer()
-                    return
+                    return self.handle_101_switching_protocols(flow)
 
                 # Upstream Proxy Mode: Handle CONNECT
                 if flow.request.first_line_format == "authority" and flow.response.status_code == 200:
                     self.handle_upstream_mode_connect(flow.request.copy())
                     return
 
-            except (ProtocolException, NetlibException) as e:
+            except (exceptions.ProtocolException, netlib.exceptions.NetlibException) as e:
                 self.send_error_response(502, repr(e))
-
                 if not flow.response:
-                    flow.error = Error(str(e))
+                    flow.error = models.Error(str(e))
                     self.channel.ask("error", flow)
-                    self.log(traceback.format_exc(), "debug")
                     return
                 else:
-                    six.reraise(ProtocolException, ProtocolException(
+                    six.reraise(exceptions.ProtocolException, exceptions.ProtocolException(
                         "Error in HTTP connection: %s" % repr(e)), sys.exc_info()[2])
             finally:
                 if flow:
                     flow.live = False
 
-    def get_request_from_client(self):
+    def get_request_from_client(self, flow):
         request = self.read_request()
+        flow.request = request
+        self.channel.ask("requestheaders", flow)
         if request.headers.get("expect", "").lower() == "100-continue":
             # TODO: We may have to use send_response_headers for HTTP2 here.
-            self.send_response(expect_continue_response)
+            self.send_response(models.expect_continue_response)
             request.headers.pop("expect")
             request.body = b"".join(self.read_request_body(request))
+            request.timestamp_end = time.time()
         return request
 
-    def send_error_response(self, code, message):
+    def send_error_response(self, code, message, headers=None):
         try:
-            response = make_error_response(code, message)
+            response = models.make_error_response(code, message, headers)
             self.send_response(response)
-        except (NetlibException, H2Error):
+        except (netlib.exceptions.NetlibException, h2.exceptions.H2Error, exceptions.Http2ProtocolException):
             self.log(traceback.format_exc(), "debug")
 
     def change_upstream_proxy_server(self, address):
@@ -257,8 +276,9 @@ class HttpLayer(Layer):
             return self.set_server(address)
 
     def handle_regular_mode_connect(self, request):
+        self.http_authenticated = True
         self.set_server((request.host, request.port))
-        self.send_response(make_connect_response(request.data.http_version))
+        self.send_response(models.make_connect_response(request.data.http_version))
         layer = self.ctx.next_layer(self)
         layer()
 
@@ -283,7 +303,7 @@ class HttpLayer(Layer):
             if callable(flow.response.stream):
                 chunks = flow.response.stream(chunks)
             self.send_response_body(flow.response, chunks)
-            flow.response.timestamp_end = utils.timestamp()
+            flow.response.timestamp_end = time.time()
 
     def get_response_from_server(self, flow):
         def get_response():
@@ -292,9 +312,9 @@ class HttpLayer(Layer):
 
         try:
             get_response()
-        except NetlibException as v:
+        except netlib.exceptions.NetlibException as e:
             self.log(
-                "server communication error: %s" % repr(v),
+                "server communication error: %s" % repr(e),
                 level="debug"
             )
             # In any case, we try to reconnect at least once. This is
@@ -308,15 +328,18 @@ class HttpLayer(Layer):
             # > server detects timeout, disconnects
             # > read (100-n)% of large request
             # > send large request upstream
+
+            if isinstance(e, exceptions.Http2ProtocolException):
+                # do not try to reconnect for HTTP2
+                raise exceptions.ProtocolException("First and only attempt to get response via HTTP2 failed.")
+
             self.disconnect()
             self.connect()
             get_response()
 
         # call the appropriate script hook - this is an opportunity for an
         # inline script to set flow.stream = True
-        flow = self.channel.ask("responseheaders", flow)
-        if flow == Kill:
-            raise Kill()
+        self.channel.ask("responseheaders", flow)
 
         if flow.response.stream:
             flow.response.data.content = None
@@ -325,7 +348,7 @@ class HttpLayer(Layer):
                 flow.request,
                 flow.response
             ))
-        flow.response.timestamp_end = utils.timestamp()
+        flow.response.timestamp_end = time.time()
 
         # no further manipulation of self.server_conn beyond this point
         # we can safely set it as the final attribute value here.
@@ -340,8 +363,7 @@ class HttpLayer(Layer):
         if self.mode == "regular":
             pass  # only absolute-form at this point, nothing to do here.
         elif self.mode == "upstream":
-            if flow.request.first_line_format == "authority":
-                flow.request.scheme = "http"  # pseudo value
+            pass
         else:
             # Setting request.host also updates the host header, which we want to preserve
             host_header = flow.request.headers.get("host", None)
@@ -350,22 +372,17 @@ class HttpLayer(Layer):
             if host_header:
                 flow.request.headers["host"] = host_header
             flow.request.scheme = "https" if self.__initial_server_tls else "http"
+        self.channel.ask("request", flow)
 
-        request_reply = self.channel.ask("request", flow)
-        if request_reply == Kill:
-            raise Kill()
-        if isinstance(request_reply, HTTPResponse):
-            flow.response = request_reply
-            return
-
-    def establish_server_connection(self, flow):
-        address = tcp.Address((flow.request.host, flow.request.port))
-        tls = (flow.request.scheme == "https")
+    def establish_server_connection(self, host, port, scheme):
+        address = tcp.Address((host, port))
+        tls = (scheme == "https")
 
         if self.mode == "regular" or self.mode == "transparent":
             # If there's an existing connection that doesn't match our expectations, kill it.
-            if address != self.server_conn.address or tls != self.server_conn.tls_established:
-                self.set_server(address, tls, address.host)
+            if address != self.server_conn.address or tls != self.server_tls:
+                self.set_server(address)
+                self.set_server_tls(tls, address.host)
             # Establish connection is neccessary.
             if not self.server_conn:
                 self.connect()
@@ -373,7 +390,7 @@ class HttpLayer(Layer):
             if not self.server_conn:
                 self.connect()
             if tls:
-                raise HttpProtocolException("Cannot change scheme in upstream proxy mode.")
+                raise exceptions.HttpProtocolException("Cannot change scheme in upstream proxy mode.")
             """
             # This is a very ugly (untested) workaround to solve a very ugly problem.
             if self.server_conn and self.server_conn.tls_established and not ssl:
@@ -391,7 +408,7 @@ class HttpLayer(Layer):
 
     def validate_request(self, request):
         if request.first_line_format == "absolute" and request.scheme != "http":
-            raise HttpException("Invalid request scheme: %s" % request.scheme)
+            raise netlib.exceptions.HttpException("Invalid request scheme: %s" % request.scheme)
 
         expected_request_forms = {
             "regular": ("authority", "absolute",),
@@ -404,7 +421,7 @@ class HttpLayer(Layer):
             err_message = "Invalid HTTP request form (expected: %s, got: %s)" % (
                 " or ".join(allowed_request_forms), request.first_line_format
             )
-            raise HttpException(err_message)
+            raise netlib.exceptions.HttpException(err_message)
 
         if self.mode == "regular" and request.first_line_format == "absolute":
             request.first_line_format = "relative"
@@ -414,10 +431,40 @@ class HttpLayer(Layer):
             if self.config.authenticator.authenticate(request.headers):
                 self.config.authenticator.clean(request.headers)
             else:
-                self.send_response(make_error_response(
-                    407,
-                    "Proxy Authentication Required",
-                    Headers(**self.config.authenticator.auth_challenge_headers())
-                ))
+                if self.mode == "transparent":
+                    self.send_response(models.make_error_response(
+                        401,
+                        "Authentication Required",
+                        http.Headers(**self.config.authenticator.auth_challenge_headers())
+                    ))
+                else:
+                    self.send_response(models.make_error_response(
+                        407,
+                        "Proxy Authentication Required",
+                        http.Headers(**self.config.authenticator.auth_challenge_headers())
+                    ))
                 return False
         return True
+
+    def handle_101_switching_protocols(self, flow):
+        """
+        Handle a successful HTTP 101 Switching Protocols Response, received after e.g. a WebSocket upgrade request.
+        """
+        # Check for WebSockets handshake
+        is_websockets = (
+            flow and
+            websockets.check_handshake(flow.request.headers) and
+            websockets.check_handshake(flow.response.headers)
+        )
+        if is_websockets and not self.config.options.websockets:
+            self.log(
+                "Client requested WebSocket connection, but the protocol is currently disabled in mitmproxy.",
+                "info"
+            )
+
+        if is_websockets and self.config.options.websockets:
+            layer = pwebsockets.WebSocketsLayer(self, flow)
+        else:
+            layer = self.ctx.next_layer(self)
+
+        layer()
