@@ -8,6 +8,10 @@ import time
 import traceback
 
 import binascii
+
+from typing import Optional  # noqa
+
+from netlib import strutils
 from six.moves import range
 
 import ssl as ssl_match_hostname
@@ -15,13 +19,14 @@ import six
 import OpenSSL
 from OpenSSL import SSL
 
-from . import certutils, version_check, utils
+from netlib import certutils
+from netlib import version_check
+from netlib import basetypes
+from netlib import exceptions
+from netlib import basethread
 
 # This is a rather hackish way to make sure that
 # the latest version of pyOpenSSL is actually installed.
-from netlib.exceptions import InvalidCertificateException, TcpReadIncomplete, TlsException, \
-    TcpTimeout, TcpDisconnect, TcpException
-
 version_check.check_pyopenssl_version()
 
 if six.PY2:
@@ -33,7 +38,7 @@ EINTR = 4
 if os.environ.get("NO_ALPN"):
     HAS_ALPN = False
 else:
-    HAS_ALPN = OpenSSL._util.lib.Cryptography_HAS_ALPN
+    HAS_ALPN = SSL._lib.Cryptography_HAS_ALPN
 
 # To enable all SSL methods use: SSLv23
 # then add options to disable certain methods
@@ -70,6 +75,7 @@ sslversion_choices = {
     "TLSv1_2": (SSL.TLSv1_2_METHOD, SSL_BASIC_OPTIONS),
 }
 
+
 class SSLKeyLogger(object):
 
     def __init__(self, filename):
@@ -160,17 +166,17 @@ class Writer(_FileLike):
 
     def flush(self):
         """
-            May raise TcpDisconnect
+            May raise exceptions.TcpDisconnect
         """
         if hasattr(self.o, "flush"):
             try:
                 self.o.flush()
             except (socket.error, IOError) as v:
-                raise TcpDisconnect(str(v))
+                raise exceptions.TcpDisconnect(str(v))
 
     def write(self, v):
         """
-            May raise TcpDisconnect
+            May raise exceptions.TcpDisconnect
         """
         if v:
             self.first_byte_timestamp = self.first_byte_timestamp or time.time()
@@ -183,7 +189,7 @@ class Writer(_FileLike):
                     self.add_log(v[:r])
                     return r
             except (SSL.Error, socket.error) as e:
-                raise TcpDisconnect(str(e))
+                raise exceptions.TcpDisconnect(str(e))
 
 
 class Reader(_FileLike):
@@ -214,17 +220,17 @@ class Reader(_FileLike):
                     time.sleep(0.1)
                     continue
                 else:
-                    raise TcpTimeout()
+                    raise exceptions.TcpTimeout()
             except socket.timeout:
-                raise TcpTimeout()
+                raise exceptions.TcpTimeout()
             except socket.error as e:
-                raise TcpDisconnect(str(e))
+                raise exceptions.TcpDisconnect(str(e))
             except SSL.SysCallError as e:
                 if e.args == (-1, 'Unexpected EOF'):
                     break
-                raise TlsException(str(e))
+                raise exceptions.TlsException(str(e))
             except SSL.Error as e:
-                raise TlsException(str(e))
+                raise exceptions.TlsException(str(e))
             self.first_byte_timestamp = self.first_byte_timestamp or time.time()
             if not data:
                 break
@@ -258,9 +264,9 @@ class Reader(_FileLike):
         result = self.read(length)
         if length != -1 and len(result) != length:
             if not result:
-                raise TcpDisconnect()
+                raise exceptions.TcpDisconnect()
             else:
-                raise TcpReadIncomplete(
+                raise exceptions.TcpReadIncomplete(
                     "Expected %s bytes, got %s" % (length, len(result))
                 )
         return result
@@ -273,7 +279,7 @@ class Reader(_FileLike):
             Up to the next N bytes if peeking is successful.
 
         Raises:
-            TcpException if there was an error with the socket
+            exceptions.TcpException if there was an error with the socket
             TlsException if there was an error with pyOpenSSL.
             NotImplementedError if the underlying file object is not a [pyOpenSSL] socket
         """
@@ -281,26 +287,17 @@ class Reader(_FileLike):
             try:
                 return self.o._sock.recv(length, socket.MSG_PEEK)
             except socket.error as e:
-                raise TcpException(repr(e))
+                raise exceptions.TcpException(repr(e))
         elif isinstance(self.o, SSL.Connection):
             try:
-                if tuple(int(x) for x in OpenSSL.__version__.split(".")[:2]) > (0, 15):
-                    return self.o.recv(length, socket.MSG_PEEK)
-                else:
-                    # TODO: remove once a new version is released
-                    # Polyfill for pyOpenSSL <= 0.15.1
-                    # Taken from https://github.com/pyca/pyopenssl/commit/1d95dea7fea03c7c0df345a5ea30c12d8a0378d2
-                    buf = SSL._ffi.new("char[]", length)
-                    result = SSL._lib.SSL_peek(self.o._ssl, buf, length)
-                    self.o._raise_ssl_error(self.o._ssl, result)
-                    return SSL._ffi.buffer(buf, result)[:]
+                return self.o.recv(length, socket.MSG_PEEK)
             except SSL.Error as e:
-                six.reraise(TlsException, TlsException(str(e)), sys.exc_info()[2])
+                six.reraise(exceptions.TlsException, exceptions.TlsException(str(e)), sys.exc_info()[2])
         else:
             raise NotImplementedError("Can only peek into (pyOpenSSL) sockets")
 
 
-class Address(utils.Serializable):
+class Address(basetypes.Serializable):
 
     """
         This class wraps an IPv4/IPv6 tuple to provide named attributes and
@@ -457,11 +454,11 @@ class _Connection(object):
     def __init__(self, connection):
         if connection:
             self.connection = connection
-            self.peer_address = Address(connection.getpeername())
+            self.ip_address = Address(connection.getpeername())
             self._makefile()
         else:
             self.connection = None
-            self.peer_address = None
+            self.ip_address = None
             self.rfile = None
             self.wfile = None
 
@@ -488,7 +485,7 @@ class _Connection(object):
                 try:
                     self.wfile.flush()
                     self.wfile.close()
-                except TcpDisconnect:
+                except exceptions.TcpDisconnect:
                     pass
 
             self.rfile.close()
@@ -508,6 +505,7 @@ class _Connection(object):
                             alpn_protos=None,
                             alpn_select=None,
                             alpn_select_callback=None,
+                            sni=None,
                             ):
         """
         Creates an SSL Context.
@@ -529,8 +527,14 @@ class _Connection(object):
         if verify_options is not None:
             def verify_cert(conn, x509, errno, err_depth, is_cert_verified):
                 if not is_cert_verified:
-                    self.ssl_verification_error = dict(errno=errno,
-                                                       depth=err_depth)
+                    self.ssl_verification_error = exceptions.InvalidCertificateException(
+                        "Certificate Verification Error for {}: {} (errno: {}, depth: {})".format(
+                            sni,
+                            strutils.native(SSL._ffi.string(SSL._lib.X509_verify_cert_error_string(errno)), "utf8"),
+                            errno,
+                            err_depth
+                        )
+                    )
                 return is_cert_verified
 
             context.set_verify(verify_options, verify_cert)
@@ -552,7 +556,7 @@ class _Connection(object):
                 # TODO: maybe change this to with newer pyOpenSSL APIs
                 context.set_tmp_ecdh(OpenSSL.crypto.get_elliptic_curve('prime256v1'))
             except SSL.Error as v:
-                raise TlsException("SSL cipher specification error: %s" % str(v))
+                raise exceptions.TlsException("SSL cipher specification error: %s" % str(v))
 
         # SSLKEYLOGFILE
         if log_ssl_key:
@@ -573,21 +577,42 @@ class _Connection(object):
             elif alpn_select_callback is not None and alpn_select is None:
                 context.set_alpn_select_callback(alpn_select_callback)
             elif alpn_select_callback is not None and alpn_select is not None:
-                raise TlsException("ALPN error: only define alpn_select (string) OR alpn_select_callback (method).")
+                raise exceptions.TlsException("ALPN error: only define alpn_select (string) OR alpn_select_callback (method).")
 
         return context
 
 
+class ConnectionCloser(object):
+    def __init__(self, conn):
+        self.conn = conn
+        self._canceled = False
+
+    def pop(self):
+        """
+            Cancel the current closer, and return a fresh one.
+        """
+        self._canceled = True
+        return ConnectionCloser(self.conn)
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args):
+        if not self._canceled:
+            self.conn.close()
+
+
 class TCPClient(_Connection):
 
-    def __init__(self, address, source_address=None):
+    def __init__(self, address, source_address=None, spoof_source_address=None):
         super(TCPClient, self).__init__(None)
         self.address = address
         self.source_address = source_address
         self.cert = None
         self.server_certs = []
-        self.ssl_verification_error = None
+        self.ssl_verification_error = None  # type: Optional[exceptions.InvalidCertificateException]
         self.sni = None
+        self.spoof_source_address = spoof_source_address
 
     @property
     def address(self):
@@ -630,7 +655,7 @@ class TCPClient(_Connection):
                 context.use_privatekey_file(cert)
                 context.use_certificate_file(cert)
             except SSL.Error as v:
-                raise TlsException("SSL client certificate error: %s" % str(v))
+                raise exceptions.TlsException("SSL client certificate error: %s" % str(v))
         return context
 
     def convert_to_ssl(self, sni=None, alpn_protos=None, **sslctx_kwargs):
@@ -644,29 +669,30 @@ class TCPClient(_Connection):
         """
         verification_mode = sslctx_kwargs.get('verify_options', None)
         if verification_mode == SSL.VERIFY_PEER and not sni:
-            raise TlsException("Cannot validate certificate hostname without SNI")
+            raise exceptions.TlsException("Cannot validate certificate hostname without SNI")
 
         context = self.create_ssl_context(
             alpn_protos=alpn_protos,
+            sni=sni,
             **sslctx_kwargs
         )
         self.connection = SSL.Connection(context, self.connection)
         if sni:
             self.sni = sni
-            self.connection.set_tlsext_host_name(sni)
+            self.connection.set_tlsext_host_name(sni.encode("idna"))
         self.connection.set_connect_state()
         try:
             self.connection.do_handshake()
         except SSL.Error as v:
             if self.ssl_verification_error:
-                raise InvalidCertificateException("SSL handshake error: %s" % repr(v))
+                raise self.ssl_verification_error
             else:
-                raise TlsException("SSL handshake error: %s" % repr(v))
+                raise exceptions.TlsException("SSL handshake error: %s" % repr(v))
         else:
             # Fix for pre v1.0 OpenSSL, which doesn't throw an exception on
             # certificate validation failure
-            if verification_mode == SSL.VERIFY_PEER and self.ssl_verification_error is not None:
-                raise InvalidCertificateException("SSL handshake error: certificate verify failed")
+            if verification_mode == SSL.VERIFY_PEER and self.ssl_verification_error:
+                raise self.ssl_verification_error
 
         self.cert = certutils.SSLCert(self.connection.get_peer_certificate())
 
@@ -682,33 +708,54 @@ class TCPClient(_Connection):
             if self.cert.cn:
                 crt["subject"] = [[["commonName", self.cert.cn.decode("ascii", "strict")]]]
             if sni:
-                hostname = sni.decode("ascii", "strict")
+                hostname = sni
             else:
                 hostname = "no-hostname"
             ssl_match_hostname.match_hostname(crt, hostname)
         except (ValueError, ssl_match_hostname.CertificateError) as e:
-            self.ssl_verification_error = dict(depth=0, errno="Invalid Hostname")
+            self.ssl_verification_error = exceptions.InvalidCertificateException(
+                "Certificate Verification Error for {}: {}".format(
+                    sni or repr(self.address),
+                    str(e)
+                )
+            )
             if verification_mode == SSL.VERIFY_PEER:
-                raise InvalidCertificateException("Presented certificate for {} is not valid: {}".format(sni, str(e)))
+                raise self.ssl_verification_error
 
         self.ssl_established = True
         self.rfile.set_descriptor(self.connection)
         self.wfile.set_descriptor(self.connection)
 
+    def makesocket(self):
+        # some parties (cuckoo sandbox) need to hook this
+        return socket.socket(self.address.family, socket.SOCK_STREAM)
+
     def connect(self):
         try:
-            connection = socket.socket(self.address.family, socket.SOCK_STREAM)
+            connection = self.makesocket()
+
+            if self.spoof_source_address:
+                try:
+                    # 19 is `IP_TRANSPARENT`, which is only available on Python 3.3+ on some OSes
+                    if not connection.getsockopt(socket.SOL_IP, 19):
+                        connection.setsockopt(socket.SOL_IP, 19, 1)
+                except socket.error as e:
+                    raise exceptions.TcpException(
+                        "Failed to spoof the source address: " + e.strerror
+                    )
             if self.source_address:
                 connection.bind(self.source_address())
             connection.connect(self.address())
             self.source_address = Address(connection.getsockname())
         except (socket.error, IOError) as err:
-            raise TcpException(
+            raise exceptions.TcpException(
                 'Error connecting to "%s": %s' %
-                (self.address.host, err))
+                (self.address.host, err)
+            )
         self.connection = connection
-        self.peer_address = Address(connection.getpeername())
+        self.ip_address = Address(connection.getpeername())
         self._makefile()
+        return ConnectionCloser(self)
 
     def settimeout(self, n):
         self.connection.settimeout(n)
@@ -769,7 +816,7 @@ class BaseHandler(_Connection):
             until then we're conservative.
         """
 
-        context = self._create_ssl_context(**sslctx_kwargs)
+        context = self._create_ssl_context(ca_pemfile=chain_file, **sslctx_kwargs)
 
         context.use_privatekey(key)
         if isinstance(cert, certutils.SSLCert):
@@ -792,10 +839,6 @@ class BaseHandler(_Connection):
                 return True
             context.set_verify(SSL.VERIFY_PEER, save_cert)
 
-        # Cert Verify
-        if chain_file:
-            context.load_verify_locations(chain_file)
-
         if dhparams:
             SSL._lib.SSL_CTX_set_tmp_dh(context._context, dhparams)
 
@@ -816,7 +859,7 @@ class BaseHandler(_Connection):
         try:
             self.connection.do_handshake()
         except SSL.Error as v:
-            raise TlsException("SSL handshake error: %s" % repr(v))
+            raise exceptions.TlsException("SSL handshake error: %s" % repr(v))
         self.ssl_established = True
         self.rfile.set_descriptor(self.connection)
         self.wfile.set_descriptor(self.connection)
@@ -834,6 +877,25 @@ class BaseHandler(_Connection):
             return b""
 
 
+class Counter:
+    def __init__(self):
+        self._count = 0
+        self._lock = threading.Lock()
+
+    @property
+    def count(self):
+        with self._lock:
+            return self._count
+
+    def __enter__(self):
+        with self._lock:
+            self._count += 1
+
+    def __exit__(self, *args):
+        with self._lock:
+            self._count -= 1
+
+
 class TCPServer(object):
     request_queue_size = 20
 
@@ -846,15 +908,17 @@ class TCPServer(object):
         self.socket.bind(self.address())
         self.address = Address.wrap(self.socket.getsockname())
         self.socket.listen(self.request_queue_size)
+        self.handler_counter = Counter()
 
     def connection_thread(self, connection, client_address):
-        client_address = Address(client_address)
-        try:
-            self.handle_client_connection(connection, client_address)
-        except:
-            self.handle_error(connection, client_address)
-        finally:
-            close_socket(connection)
+        with self.handler_counter:
+            client_address = Address(client_address)
+            try:
+                self.handle_client_connection(connection, client_address)
+            except:
+                self.handle_error(connection, client_address)
+            finally:
+                close_socket(connection)
 
     def serve_forever(self, poll_interval=0.1):
         self.__is_shut_down.clear()
@@ -870,12 +934,16 @@ class TCPServer(object):
                         raise
                 if self.socket in r:
                     connection, client_address = self.socket.accept()
-                    t = threading.Thread(
+                    t = basethread.BaseThread(
+                        "TCPConnectionHandler (%s: %s:%s -> %s:%s)" % (
+                            self.__class__.__name__,
+                            client_address[0],
+                            client_address[1],
+                            self.address.host,
+                            self.address.port
+                        ),
                         target=self.connection_thread,
                         args=(connection, client_address),
-                        name="ConnectionThread (%s:%s -> %s:%s)" %
-                             (client_address[0], client_address[1],
-                              self.address.host, self.address.port)
                     )
                     t.setDaemon(1)
                     try:
@@ -899,7 +967,7 @@ class TCPServer(object):
         """
         # If a thread has persisted after interpreter exit, the module might be
         # none.
-        if traceback:
+        if traceback and six:
             exc = six.text_type(traceback.format_exc())
             print(u'-' * 40, file=fp)
             print(
@@ -917,3 +985,14 @@ class TCPServer(object):
         """
             Called after server shutdown.
         """
+
+    def wait_for_silence(self, timeout=5):
+        start = time.time()
+        while 1:
+            if time.time() - start >= timeout:
+                raise exceptions.Timeout(
+                    "%s service threads still alive" %
+                    self.handler_counter.count
+                )
+            if self.handler_counter.count == 0:
+                return
